package rob.math.geom;


import javax.swing.JOptionPane;
import rob.util.*;
import rob.exceptions.*;
import java.util.Collection;
import java.util.List;
import java.util.ListIterator;
import java.util.ConcurrentModificationException;
import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.PathIterator;
import java.awt.geom.AffineTransform;
import java.awt.Rectangle;
import java.awt.Point;


public class CDLLPolygon extends AbstractCDLList implements IPolygon
{
	protected double xMin, xMax, yMin, yMax;	// Boundingbox of the polygon.
	protected boolean hasConvex = false;	// speeds up some operations.
	protected boolean isConvex = false;		// speeds up some operations.
	protected boolean hasArea = false;		// speeds up some operations.
	protected double area = 0;				// speeds up some operations.
	protected boolean boundsModified = false;	// speeds up some operations.


	public CDLLPolygon()	{super();}


	public CDLLPolygon(Collection points)
	{
		super(points);
		if (!this.isPolygon())
			throw new InvalidParamException("Not a polyigon");
		calculateBounds();
	}


	public CDLLPolygon(Rectangle2D rect)
	{
		super();
		xMin = rect.getMinX(); yMin = rect.getMinY();
		xMax = rect.getMaxX(); yMax = rect.getMaxY();
		hasConvex = true; isConvex = true; hasArea = true;
		area = (xMax - xMin)*(yMax - yMin);
		super.add(new Point2D.Double(xMin, yMin));
		super.add(new Point2D.Double(xMin, yMax));
		super.add(new Point2D.Double(xMax, yMax));
		super.add(new Point2D.Double(xMax, yMin));
	}


	public CDLLPolygon(Shape shp)
	{
		this();
 		PathIterator itShp = null;
 		itShp = shp.getPathIterator(null, 1);
		float [] tmpPoint = new float[6];
		int retValue = 0;
		while (!itShp.isDone())
		{
			retValue = itShp.currentSegment(tmpPoint);
			if (retValue != PathIterator.SEG_CLOSE)
				super.add(new Point2D.Double(tmpPoint[0], tmpPoint[1]));
			itShp.next();
		}
	}


/**
 *	Consome tempo O(n).
 */
	public int setVertexAsFirst(int index)
	{
		moveHead(index);
		return 1;
	}


/**
 *	Consome tempo O(n).
 */
	public int remove(int fromIndex, int toIndex)
	{
		Entry saveHead = head;
		fromIndex = moveHead(fromIndex);
		toIndex = adjustToPositiveInBoundsValue(toIndex);
		toIndex = (toIndex < fromIndex) ? fromIndex - toIndex : toIndex - fromIndex;
		while (--toIndex > -1)
			remove(head);
		head = saveHead;
		return 1;
	}


/**
 *	Consome tempo O(1).
 */
	public Point2D getPoint(int index)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		return (Point2D) saveHead.element;
	}


/**
 *	Consome tempo O(1).
 */
	public double getX(int index)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		return ((Point2D) saveHead.element).getX();
	}


/**
 *	Consome tempo O(1).
 */
	public double getY(int index)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		return ((Point2D) saveHead.element).getY();
	}


/**
 *	Consome tempo O(n).
 */
	public void setPoint(int index, Point2D newPoint)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		Point2D savePoint = (Point2D) saveHead.element;
		if (intersectsIfNotEntry(saveHead, newPoint, savePoint))
			throw new InvalidParamException("Wouldn't be a polyigon");
		else
		{
			saveHead.element = newPoint;
			boundsModified = true;
		}
	}


/**
 *	Consome tempo O(n).
 */
	public void setX(int index, double newX)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		Point2D savePoint = (Point2D) saveHead.element;
		Point2D newPoint = new Point((int) newX, (int) savePoint.getY());
		if (intersectsIfNotEntry(saveHead, newPoint, savePoint))
			throw new InvalidParamException("Wouldn't be a polyigon");
		else
		{
			saveHead.element = newPoint;
			boundsModified = true;
		}
	}


/**
 *	Consome tempo O(n).
 */
	public void setY(int index, double newY)
	{
		if (head == null)
			throw new NoSuchElementException("Empty polygon.");
		Entry saveHead = head;
		saveHead = moveEntry(index, saveHead);
		Point2D savePoint = (Point2D) saveHead.element;
		Point2D newPoint = new Point((int) savePoint.getX(), (int) newY);
		if (intersectsIfNotEntry(saveHead, newPoint, savePoint))
			throw new InvalidParamException("Wouldn't be a polyigon");
		else
		{
			saveHead.element = newPoint;
			boundsModified = true;
		}
	}


	public IPolygonIterator polygonIterator()
	{
		return CDLLPolygon.this.new CDLLPolygonIterator();
	}


/**
 *	Consome tempo O(n).
 */
	public int getSubPolygons(int fromIndex, int toIndex, IPolygon onePolygon, IPolygon otherPolygon)
	{
		if (!isDiagonal(fromIndex, toIndex))
			throw new InvalidParamException("Wouldn't be polyigons");
		onePolygon.clear();
		otherPolygon.clear();
		ListIterator itList = listIterator();
		int i = - 1;
		while (++i < fromIndex)
			otherPolygon.add(itList.next());
		itList.previous();
		i--;
		while (++i < toIndex)
			onePolygon.add(itList.next());
		itList.previous();
		i--;
		while (++i < size)
			otherPolygon.add(itList.next());
		return 1;
	}


/**
 *	Consome tempo O(n).
 */
	public int split(int fromIndex, int toIndex, IPolygon otherPolygon)
	{
		if (!isDiagonal(fromIndex, toIndex))
			throw new InvalidParamException("Wouldn't be polyigons");
		Entry new1 = moveEntry(head, fromIndex);
		toIndex = adjustToPositiveInBoundsValue(toIndex);
		toIndex = (toIndex < fromIndex) ? size - (fromIndex - toIndex) : toIndex - fromIndex;
		int moveToFrom = size - toIndex;
		Entry new2 = moveEntry(new1, toIndex);

		Entry saveHead = head;
		head = new2;
		ListIterator itList = listIterator();
		int i = - 1;
		while (++i < moveToFrom)
			otherPolygon.add(itList.next());
		head = saveHead;
		
		new2.next = new1;
		new1.previous = new2;
		return 1;
	}


/**
 *	Consome tempo O(n).
 */
	public int split(int fromIndex, int toIndex)
	{
		if (!isDiagonal(fromIndex, toIndex))
			throw new InvalidParamException("Wouldn't be polyigons");
		Entry new1 = moveEntry(head, fromIndex);
		toIndex = adjustToPositiveInBoundsValue(toIndex);
		Entry new2 = moveEntry(new1, (toIndex < fromIndex) ? size - (fromIndex - toIndex) : toIndex - fromIndex);
		new2.next = new1;
		new1.previous = new2;
		return 1;
	}


/**
 *	Consome tempo O(n).
 */
	public int split(int toIndex)
	{
		if (!isDiagonal(0, toIndex))
			throw new InvalidParamException("Wouldn't be polyigons");
		Entry new1 = head;
		toIndex = adjustToPositiveInBoundsValue(toIndex);
		if (toIndex < 0)
			toIndex += size;
		Entry new2 = moveEntry(new1, toIndex);
		new2.next = new1;
		new1.previous = new2;
		return 1;
	}



/**
 *	Esta operao requer que o polgono esteje orientado no sentido anti-horrio. Consome tempo O(n) e espao O(1).
 */
	public boolean isDiagonal(int index1, int index2)
	{
		if (!insideCone(index1, index2))
			return false;

		index1 = adjustToPositiveInBoundsValue(index1);
		index2 = adjustToPositiveInBoundsValue(index2);
		Entry pointer1 = moveEntry(index1, head);
		Entry pointer2 = moveEntry(index2, head);
		Point2D d1 = (Point2D) pointer1.element;
		Point2D d2 = (Point2D) pointer2.element;
		Entry saveHead = head;
		do
		{
			if (saveHead == pointer1 || saveHead == pointer2 || saveHead.next == pointer1 || saveHead.next == pointer2)
			{
				saveHead = saveHead.next;
				continue;
			}
			if (GeomOperations.intersect(d1, d2, (Point2D) saveHead.element, (Point2D) saveHead.next.element))
			{
				System.out.println("Cruza o lado! (" + ((Point2D) saveHead.element).getX() + ", " + ((Point2D) saveHead.element).getY() + ")");
				return false;
			}
			saveHead = saveHead.next;
		}
		while (saveHead != head);
		return true;
	}


/**
 *	Esta operao requer que o polgono esteje orientado no sentido anti-horrio. Se ele no estiver, ela o reorienta e 
 * leva tempo O(n), seno, leva tempo O(1).
 */
	public double getArea()
	{
		if (hasArea)
			return area;
		hasArea = true;
		area = 0;
		Entry saveHead = head;
		do
		{
			area += ((Point2D) saveHead.element).getX()*((Point2D) saveHead.next.element).getY() - ((Point2D) saveHead.element).getY()*((Point2D) saveHead.next.element).getX();
			saveHead = saveHead.next;
		}
		while (saveHead != head);
		area = area/2;
		if (area < 0)
		{
			area = -area;
			reorientPolygon();
		}
		return area;
	}


/**
 *	Esta operao requer que o polgono esteje orientado no sentido anti-horrio. Leva tempo O(n).
 */
	public boolean isConvex()
	{
		if (hasConvex)
			return isConvex;
		hasConvex = true;
		Point2D p1 = null;
		Point2D p2 = null;
		Point2D p3 = null;
		Entry saveHead = head;
		p2 = (Point2D) saveHead.previous.previous.element;
		p3 = (Point2D) saveHead.previous.element;
		do
		{
			p1 = p2; p2 = p3;
			p3 = (Point2D) saveHead.element;
			if (!GeomOperations.isLeftOrOn(p1, p2, p3))
			{
				isConvex = false;
				return false;
			}
			saveHead = saveHead.next;
		}
		while (saveHead != head);
		isConvex = true;
		return true;
	}


/**
 *	Consome tempo O(index).
 */
	public boolean isConvex(int index)
	{
		Entry tmpHead = moveEntry(head, index);
		return GeomOperations.isLeftOrOn((Point2D) tmpHead.previous.element, (Point2D) tmpHead.element, (Point2D) tmpHead.next.element);
	}


/**
 *	Consome tempo O(n*lg(n)).
 */
	public boolean isPolygon()
	{
		if ((new SegmentsCrossings(this)).hasCrossings())
			return false;
		if (getArea() < 0)
			reorientPolygon();
		return true;
	}


/**
 *	Consome tempo O(n).
 */
	public boolean intersects(Point2D p1, Point2D p2)
	{
		Entry tmp = head;
		do
		{
			if (GeomOperations.intersect(p1, p2, (Point2D) tmp.element, (Point2D) tmp.next.element))
				return true;
			tmp = tmp.next;
		}
		while (tmp != head);
		return false;
	}


/**
 *	As operaes de insero e remoo com o Iterator so isentas de verificao. 
 * Assim, todas operaes so de ordem O(1).
 */
	public class CDLLPolygonIterator extends CDLListIterator implements IPolygonIterator
	{
		public CDLLPolygonIterator()		{}


		public Object next()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			if (!hasNext())
				return null;
			head2 = head2.next;
			itIndex++;
			lastReturned = head2;
			return lastReturned.element;
		}


		public Object previous()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			if (!hasPrevious())
				return null;
			itIndex--;
			lastReturned = head2;
			head2 = head2.previous;
			return lastReturned.element;
		}


/**
 *	Adiciona logo aps o ltimo elemento retornado, ou seja, ser o prximo a ser retornado por next().
 */
		public int addAfter(Point2D newPoint)
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			CDLLPolygon.this.addAfter(lastReturned, newPoint);
			localModificationCounter++;
			return itIndex;
		}


/**
 *	Adiciona logo antes do ltimo elemento retornado, ou seja, ser o prximo a ser retornado por previous().
 * Faz o mesmo que o add normal da interface ListIterator s que com um nome decente.
 */
		public int addBefore(Point2D newPoint)
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			CDLLPolygon.this.addBefore(lastReturned, newPoint);
			localModificationCounter++;			
			return ++itIndex;
		}


/**
 *	Remove o Point2D logo antes do ltimo elemento retornado.
 */
		public int removeNext()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			CDLLPolygon.this.remove(lastReturned.next);
			localModificationCounter++;			
			return itIndex;
		}


/**
 *	Remove o Point2D logo antes do ltimo elemento retornado.
 */
		public int removePrevious()
		{
			if (wasModified())
				throw new ConcurrentModificationException();
			CDLLPolygon.this.remove(lastReturned.previous);
			localModificationCounter++;			
			return --itIndex;
		}


		public Point2D getPoint()	{return (Point2D) lastReturned.element;}
		public double getX()	{return ((Point2D) lastReturned.element).getX();}
		public double getY()	{return ((Point2D) lastReturned.element).getY();}
		public  void setPoint(Point2D newPoint)	{lastReturned.element = newPoint;}
		public void setX(double newX)	{((Point2D) lastReturned.element).setLocation(newX, ((Point2D) lastReturned.element).getY());}
		public void setY(double newY)	{((Point2D) lastReturned.element).setLocation(((Point2D) lastReturned.element).getX(), newY);}
	}


// Operaes internas do polgono.

/**
 *	Consome tempo O(n).
 */
	protected void calculateBounds()
	{
		ListIterator polygonIterator = listIterator();
		Point2D pTmp = null;
		if (polygonIterator.hasNext())
		{
			pTmp = (Point2D) polygonIterator.next();
			xMin = pTmp.getX();
			xMax = xMin;
			yMin = pTmp.getY();
			yMax = yMin;
		}
		while (polygonIterator.hasNext())
		{
			pTmp = (Point2D) polygonIterator.next();
			if (pTmp.getX() < xMin)
				xMin = pTmp.getX();
			else
				if (pTmp.getX() > xMax)
					xMax = pTmp.getX();
			if (pTmp.getY() < yMin)
				yMin = pTmp.getY();
			else
				if (pTmp.getY() > yMax)
					yMax = pTmp.getY();
		}
		boundsModified = false;
	}


/**
 *	Consome tempo O(1).
 */
	protected double area2(int indexFirst, Point2D pTest)
	{
		Entry tmpHead = moveEntry(head, indexFirst);
		Point2D p1 = (Point2D) tmpHead.element;
		Point2D p2 = (Point2D) tmpHead.next.element;
		return GeomOperations.area2(p1, p2, pTest);
	}


/**
 *	Consome tempo O(1).
 */
	protected boolean isLeft(int indexFirst, Point2D pTest)
	{
		Entry tmpHead = moveEntry(head, indexFirst);
		Point2D p1 = (Point2D) tmpHead.element;
		Point2D p2 = (Point2D) tmpHead.next.element;
		return GeomOperations.isLeft(p1, p2, pTest);
	}


/**
 *	Testa se o vrtice j se encontra dentro do cone determinado pelo vrtice i e pelas semi-retas que comeam no
 * vrtice i e contm as arestas incidentes em i. Consome tempo O(1).
 */
	protected boolean insideCone(int i, int j)
	{
		i = adjustToPositiveInBoundsValue(i);
		j = adjustToPositiveInBoundsValue(j);
		int iAnt  = (i - 1 + size)%size;
		int iProx = (i + 1)%size;
		Entry saveHead = moveEntry(iAnt, head);
		Entry loadHead = moveEntry(j, head);
		if (GeomOperations.isLeftOrOn((Point2D) saveHead.element, (Point2D) saveHead.next.element, (Point2D) saveHead.next.next.element))
			return GeomOperations.isLeft((Point2D) saveHead.next.element, (Point2D) loadHead.element, (Point2D) saveHead.element)
			    && GeomOperations.isLeft((Point2D) loadHead.element, (Point2D) saveHead.next.element, (Point2D) saveHead.next.next.element);
		else
			return !(GeomOperations.isLeftOrOn((Point2D) saveHead.next.element, (Point2D) loadHead.element, (Point2D) saveHead.next.next.element)
                        && GeomOperations.isLeftOrOn((Point2D) loadHead.element, (Point2D) saveHead.next.element, (Point2D) saveHead.element));
	}


/**
 *	Consome tempo O(n).
 */
	protected int reorientPolygon()
	{
		Entry tmpEntry = null;
		Entry saveHead = head;
		do
		{
			tmpEntry = saveHead.next; saveHead.next = saveHead.previous; saveHead.previous = tmpEntry;
			saveHead = saveHead.previous;
		}
		while (saveHead != head);
		return 1;
	}


/**
 *	Consome tempo O(n).
 */
	protected boolean intersectsIfNotEntry(Entry entry, Point2D p1, Point2D p2)
	{
		Entry saveHead = head;
		do
		{
			if (saveHead != entry && saveHead.next != entry && GeomOperations.intersect(p1, p2, (Point2D) saveHead.element, (Point2D) saveHead.next.element))
				return true;
			saveHead = saveHead.next;
		}
		while (saveHead != head);
		return false;
	}
}