/*
 * @(#)TwoPolygonsCrossings.java	1.0 2000/07/22
 *
 */
package rob.math.geom;

import java.awt.Color;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.util.Collection;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import rob.math.*;
import rob.util.*;
import rob.exceptions.*;


/**
 * Um objeto <code>TwoPolygonsCrossings</code> recebe como parmetros dois
 * polgonos e pode extrair deles os pontos de interseces, o polgono dado
 * pela unio, e o polgono dado pela interseco.
 * O algoritmo usado demora tempo proporcional a (n + k)*lg(n), onde n  o nmero
 * de vrtices somados dos polgonos e k  o nmero de cruzamentos entre as 
 * arestas dos polgonos.
 * O algoritmo no garante resposta nem que funcione para alguns casos especiais
 * listados a seguir:
 *	1- Vrtices coincidentes.
 *	2- Arestas com segmentos coincidentes.
 *	3- Entradas que no representem polgonos.
 *	4- Caso o objeto de retorno no seja um polgono.
 * <p>
 *
 * @version 	1.0 2000/07/22
 * @author 	Roberto Giordano Barra
 * @see         
 */
public class TwoPolygonsCrossings
{
	// As constantes abaixo representam as regies que podem se formar quando os polgonos dividem
	// o plano em que se encontram. A regio de interseco  dada por A + B.
	public static final byte VACUUM = 0;
	public static final byte A = 1;
	public static final byte B = 2;

	// Guarda a fila de eventos.
	TwoPolygonsEventStruct eventStruct = null;

	// Guarda a linha de varredura.
	TwoPolygonsSweepLine sweepLine = null;

	// Guardam os polgonos examinados.
	IPolygon polyA = null, polyB = null;


	/**
	 * Cria um novo objeto que procura as interseces entre os polgonos polyA e polyB.
	 */
	public TwoPolygonsCrossings(IPolygon polyA, IPolygon polyB)
	{
		this.polyA = polyA;
		this.polyB = polyB;
	}	


	/**
	 * Cria e estrutura que conter os pontos-eventos (ordenados em ordem crescente primeiro em x
	 * e depois em y) e adiciona os pontos dos polgonos nela.
	 */
	public void createEventStruct()
	{
		eventStruct = new TwoPolygonsEventStruct();

		// Adiciona polgono A.
		Iterator tmpIt = polyA.iterator();
		Point2D p1 = null, p2 = null;
		int i = -1;
		if (tmpIt.hasNext())
			p2 = (Point2D) tmpIt.next();
		while (tmpIt.hasNext())
		{
			p1 = p2; p2 = (Point2D) tmpIt.next();
			eventStruct.insertSegment(p1, p2, TwoPolygonsCrossings.A);
		}
		eventStruct.insertSegment(p2, (Point2D) polyA.get(0), TwoPolygonsCrossings.A);

		// Adiciona polgono B.
		tmpIt = polyB.iterator();
		p1 = null; p2 = null;
		i = -1;
		if (tmpIt.hasNext())
			p2 = (Point2D) tmpIt.next();
		while (tmpIt.hasNext())
		{
			p1 = p2; p2 = (Point2D) tmpIt.next();
			eventStruct.insertSegment(p1, p2, TwoPolygonsCrossings.B);
		}
		eventStruct.insertSegment(p2, (Point2D) polyB.get(0), TwoPolygonsCrossings.B);
	}	


	/**
	 * Calcula a unio entre os polgonos A e B em tempo proporcional (n + k)*lg(n), onde k  o
	 * nmero de interseces e n o nmero de pontos da amostra.
	 * @return  lista contendo os pontos em sentido anti-horrio que formam o polgono da unio entre A e B.
	 */
	public List getUnion()
	{
		createEventStruct();
		TwoPolygonsOutputGenerator outputGenerator = new TwoPolygonsOutputGenerator(TwoPolygonsOutputGenerator.UNION);
		TwoPolygonsRegionsMarkersTerminatorFactory terminatorFactory = new TwoPolygonsRegionsMarkersTerminatorFactory();
		sweepLine = new TwoPolygonsSweepLine(terminatorFactory);
		AVLTerminatedTree.AVLNode actualEvent = eventStruct.getNextAndDelete();
		while (actualEvent != null)
		{
			terminatorFactory.stepTerminator(actualEvent);
			stepSweepLine(actualEvent);
			outputGenerator.stepOutput(actualEvent);
			actualEvent = eventStruct.getNextAndDelete();
		}
		return (List) outputGenerator.getAnswer();
	}
	

	/**
	 * Calcula a interseco entre os polgonos A e B em tempo proporcional (n + k)*lg(n), onde k  o
	 * nmero de interseces e n o nmero de pontos da amostra.
	 * @return  lista contendo listas contendo os pontos em sentido anti-horrio que formam o polgono 
	 * da interseco entre A e B.
	 */
	public List getIntersection()
	{
		createEventStruct();
		TwoPolygonsOutputGenerator outputGenerator = new TwoPolygonsOutputGenerator(TwoPolygonsOutputGenerator.INTERSECTION);
		TwoPolygonsRegionsMarkersTerminatorFactory terminatorFactory = new TwoPolygonsRegionsMarkersTerminatorFactory();
		sweepLine = new TwoPolygonsSweepLine(terminatorFactory);
		AVLTerminatedTree.AVLNode actualEvent = eventStruct.getNextAndDelete();
		while (actualEvent != null)
		{
			terminatorFactory.stepTerminator(actualEvent);
			stepSweepLine(actualEvent);
			outputGenerator.stepOutput(actualEvent);
			actualEvent = eventStruct.getNextAndDelete();
		}
		return (List) outputGenerator.getAnswer();
	}
	

	/**
	 * Atualiza a linha de varredura com o ltimo evento retirado da fila de eventos. Esse
	 * mtodo  realizado em tempo lg(n).
	 * Nessa verso do mtodo,  passado um objeto Graphics2D para onde ser produzida informao de
	 * debug.
	 * @param   g	o objeto Graphics2D onde se deseja plotar a informao de debug.
	 *		i	um contador para auxiliar no preparo da sada de debug.
	 *		actualEvent	o ltimo evento retirado da fila de eventos.
	 */
	public void stepSweepLineDebug(AVLTerminatedTree.AVLNode actualEvent, Graphics2D g, int i)
	{
		TwoPolygonsEventStruct.EventStructKey actualEventStructKey = (TwoPolygonsEventStruct.EventStructKey) actualEvent.key;
		AVLRightTerminatedTree.AVLNode actualSweepLineEvent = null;
		TwoPolygonsSweepLine.SweepLineKey actualSweepLineKey = null;
		String s = null;
		if (actualEventStructKey.type == TwoPolygonsEventStruct.BEGIN)
		{
			s = "BEGIN:";
			actualSweepLineEvent = actualEventStructKey.sweepNode;
			actualSweepLineKey = (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key;
			sweepLine.insert(actualSweepLineEvent);
			AVLRightTerminatedTree.AVLNode lowerSweepLineEvent = sweepLine.getAntecessor(actualSweepLineEvent);
			if (lowerSweepLineEvent != null)
			{
				s += " lowerSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) lowerSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) lowerSweepLineEvent.key).p1.getY() + ")";
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerSweepLineEvent, actualSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
			s += " actualSweepLineEvent:(" + (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key + ")";
			AVLRightTerminatedTree.AVLNode upperSweepLineEvent = sweepLine.getSuccessor(actualSweepLineEvent);
			if (upperSweepLineEvent != null)
			{
				s += " upperSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) upperSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) upperSweepLineEvent.key).p1.getY() + ")";
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(actualSweepLineEvent, upperSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
			g.drawString(s, 0, i*15);
		}
		else if (actualEventStructKey.type == TwoPolygonsEventStruct.END)
		{
			s = "END:";
			actualSweepLineEvent = actualEventStructKey.sweepNode;
			actualSweepLineKey = (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key;
			AVLRightTerminatedTree.AVLNode lowerSweepLineEvent = sweepLine.getAntecessor(actualSweepLineEvent);
			AVLRightTerminatedTree.AVLNode upperSweepLineEvent = sweepLine.getSuccessor(actualSweepLineEvent);
			sweepLine.delete(actualSweepLineEvent);
			if (lowerSweepLineEvent != null && upperSweepLineEvent != null)
			{
				s += " lowerSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) lowerSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) lowerSweepLineEvent.key).p1.getY() + ")";
				s += " actualSweepLineEvent:(" + (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key + ")";
				s += " upperSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) upperSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) upperSweepLineEvent.key).p1.getY() + ")";
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerSweepLineEvent, upperSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
			else
				s += " actualSweepLineEvent:(" + (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key + ")";
			g.drawString(s, 0, i*15);
		}
		else // (actualEventStructKey.type == TwoPolygonsEventStruct.CROSS)
		{
			s = "CROSS:";
			g.drawOval((int) actualEventStructKey.pEvt.getX() - 1, (int) actualEventStructKey.pEvt.getY() - 1, 2, 2);
			TwoPolygonsEventStruct.EventCrossKey actualEventCrossKey = (TwoPolygonsEventStruct.EventCrossKey) actualEventStructKey;

			// swap  conferido pois podem ocorrer casos quando a interseco 
			// imprpria ou muito prxima das extremidades onde o valor do ponto de interseco
			//  calculado como sendo um pouco maior do que realmente , fazendo o ponto de interseco
			// aparecer na TwoPolygonsEventStruct depois que um dos segmentos que compe a interseso j tiver
			// sido eliminado da TwoPolygonsSweepLine, ou aparecer antes de tal segemnto entrar na TwoPolygonsSweepLine. Caso
			// ocorra um dos dois casos, swap detecta que um dos segmentos no est na estrutura
			// e retorna false, impendindo que segmentos que nada tenham a ver sejam invertidos destruindo a
			// estrutura. Com essa tcnica pode-se perder alguns pontos de interseco, mas evita-se loops 
			// inifitos. Para vermos que isso no afeta a corretude do algoritmo, basta verificarmos que essa
			// tcnica equivale a encolher um pouco uma das retas, fazendo com que ele se extingue (ou comece)
			// um pouco antes (ou pouco depois) da interseco ocorrer.

			if (sweepLine.swap(actualEventCrossKey.antecessorNode, actualEventCrossKey.successorNode))
			{
				// Agora, antecessorNode est acima de successorNode.
				AVLRightTerminatedTree.AVLNode lowerLowerSweepLineEvent = sweepLine.getAntecessor(actualEventCrossKey.successorNode);
				if (lowerLowerSweepLineEvent != null)
				{
					s += " lowerLowerSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) lowerLowerSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) lowerLowerSweepLineEvent.key).p1.getY() + ")";
					AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerLowerSweepLineEvent, actualEventCrossKey.successorNode);
					if (possibleCrossEvent != null)
						eventStruct.insertEvent(possibleCrossEvent);
				}
				s += " successorNode:(" + ((TwoPolygonsSweepLine.SweepLineKey) actualEventCrossKey.successorNode.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) actualEventCrossKey.successorNode.key).p1.getY() + ")";
				s += " antecessorNode:(" + ((TwoPolygonsSweepLine.SweepLineKey) actualEventCrossKey.antecessorNode.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) actualEventCrossKey.antecessorNode.key).p1.getY() + ")";
				AVLRightTerminatedTree.AVLNode upperUpperSweepLineEvent = sweepLine.getSuccessor(actualEventCrossKey.antecessorNode);
				if (upperUpperSweepLineEvent != null)
				{
					s += " upperUpperSweepLineEvent:(" + ((TwoPolygonsSweepLine.SweepLineKey) upperUpperSweepLineEvent.key).p1.getX() + ", " + ((TwoPolygonsSweepLine.SweepLineKey) upperUpperSweepLineEvent.key).p1.getY() + ")";
					AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(actualEventCrossKey.antecessorNode, upperUpperSweepLineEvent);
					if (possibleCrossEvent != null)
						eventStruct.insertEvent(possibleCrossEvent);
				}
			}
			g.drawString(s, 0, i*15);
		}
	}


	/**
	 * Atualiza a linha de varredura com o ltimo evento retirado da fila de eventos. Esse
	 * mtodo  realizado em tempo lg(n).
	 * @param   actualEvent	o ltimo evento retirado da fila de eventos.
	 */
	public void stepSweepLine(AVLTerminatedTree.AVLNode actualEvent)
	{
		TwoPolygonsEventStruct.EventStructKey actualEventStructKey = (TwoPolygonsEventStruct.EventStructKey) actualEvent.key;
		AVLRightTerminatedTree.AVLNode actualSweepLineEvent = null;
		TwoPolygonsSweepLine.SweepLineKey actualSweepLineKey = null;
		if (actualEventStructKey.type == TwoPolygonsEventStruct.BEGIN)
		{
			actualSweepLineEvent = actualEventStructKey.sweepNode;
			actualSweepLineKey = (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key;
			sweepLine.insert(actualSweepLineEvent);
			AVLRightTerminatedTree.AVLNode lowerSweepLineEvent = sweepLine.getAntecessor(actualSweepLineEvent);
			if (lowerSweepLineEvent != null)
			{
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerSweepLineEvent, actualSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
			AVLRightTerminatedTree.AVLNode upperSweepLineEvent = sweepLine.getSuccessor(actualSweepLineEvent);
			if (upperSweepLineEvent != null)
			{
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(actualSweepLineEvent, upperSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
		}
		else if (actualEventStructKey.type == TwoPolygonsEventStruct.END)
		{
			actualSweepLineEvent = actualEventStructKey.sweepNode;
			actualSweepLineKey = (TwoPolygonsSweepLine.SweepLineKey) actualSweepLineEvent.key;
			AVLRightTerminatedTree.AVLNode lowerSweepLineEvent = sweepLine.getAntecessor(actualSweepLineEvent);
			AVLRightTerminatedTree.AVLNode upperSweepLineEvent = sweepLine.getSuccessor(actualSweepLineEvent);
			sweepLine.delete(actualSweepLineEvent);
			if (lowerSweepLineEvent != null && upperSweepLineEvent != null)
			{
				AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerSweepLineEvent, upperSweepLineEvent);
				if (possibleCrossEvent != null)
					eventStruct.insertEvent(possibleCrossEvent);
			}
		}
		else // (actualEventStructKey.type == TwoPolygonsEventStruct.CROSS)
		{
			TwoPolygonsEventStruct.EventCrossKey actualEventCrossKey = (TwoPolygonsEventStruct.EventCrossKey) actualEventStructKey;

			// swap  conferido pois podem ocorrer casos quando a interseco 
			// imprpria ou muito prxima das extremidades onde o valor do ponto de interseco
			//  calculado como sendo um pouco maior do que realmente , fazendo o ponto de interseco
			// aparecer na TwoPolygonsEventStruct depois que um dos segmentos que compe a interseso j tiver
			// sido eliminado da TwoPolygonsSweepLine, ou aparecer antes de tal segemnto entrar na TwoPolygonsSweepLine. Caso
			// ocorra um dos dois casos, swap detecta que um dos segmentos no est na estrutura
			// e retorna false, impendindo que segmentos que nada tenham a ver sejam invertidos destruindo a
			// estrutura. Com essa tcnica pode-se perder alguns pontos de interseco, mas evita-se loops 
			// inifitos. Para vermos que isso no afeta a corretude do algoritmo, basta verificarmos que essa
			// tcnica equivale a encolher um pouco uma das retas, fazendo com que ele se extingue (ou comece)
			// um pouco antes (ou pouco depois) da interseco ocorrer.

			if (sweepLine.swap(actualEventCrossKey.antecessorNode, actualEventCrossKey.successorNode))
			{
				// Agora, antecessorNode est acima de successorNode.
				AVLRightTerminatedTree.AVLNode lowerLowerSweepLineEvent = sweepLine.getAntecessor(actualEventCrossKey.successorNode);
				if (lowerLowerSweepLineEvent != null)
				{
					AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(lowerLowerSweepLineEvent, actualEventCrossKey.successorNode);
					if (possibleCrossEvent != null)
						eventStruct.insertEvent(possibleCrossEvent);
				}
				AVLRightTerminatedTree.AVLNode upperUpperSweepLineEvent = sweepLine.getSuccessor(actualEventCrossKey.antecessorNode);
				if (upperUpperSweepLineEvent != null)
				{
					AVLTerminatedTree.AVLNode possibleCrossEvent = getCrossEvent(actualEventCrossKey.antecessorNode, upperUpperSweepLineEvent);
					if (possibleCrossEvent != null)
						eventStruct.insertEvent(possibleCrossEvent);
				}
			}
		}
	}


	/**
	 * Calcula a interseco entre os segmentos representados pelos eventos da linha de varredura passados.
	 * mtodo  realizado em tempo O(1). 
	 * @param	successorNode	um dos eventos a ser testado.
	 *		antecessorNode	o outro evento a tomar parte do teste.
	 * @returns	o evento que foi gerado da interseco dos eventos fornecidos ou null em caso de no
	 * existir interseco.
	 */
	public AVLTerminatedTree.AVLNode getCrossEvent(AVLRightTerminatedTree.AVLNode antecessorNode, AVLRightTerminatedTree.AVLNode successorNode)
	{
		TwoPolygonsSweepLine.SweepLineKey antecessorKey = (TwoPolygonsSweepLine.SweepLineKey) antecessorNode.key;
		TwoPolygonsSweepLine.SweepLineKey successorKey = (TwoPolygonsSweepLine.SweepLineKey) successorNode.key;

		boolean intersects = GeomOperations.intersect(antecessorKey.p1, antecessorKey.p2, successorKey.p1, successorKey.p2);
		if (!intersects)
			return null;
		else
		{	
			double x, y;
			if (antecessorKey.p2.getX() - antecessorKey.p1.getX() == 0)
			{
				x = antecessorKey.p1.getX();
				y = (successorKey.p2.getY() - successorKey.p1.getY())/(successorKey.p2.getX() - successorKey.p1.getX())*(x - successorKey.p1.getX()) + successorKey.p1.getY();
			}
			else if (successorKey.p2.getX() - successorKey.p1.getX() == 0)
			{
				x = successorKey.p1.getX();
				y = (antecessorKey.p2.getY() - antecessorKey.p1.getY())/(antecessorKey.p2.getX() - antecessorKey.p1.getX())*(x - antecessorKey.p1.getX()) + antecessorKey.p1.getY();
			}
			else
			{	
				double m1 = (antecessorKey.p2.getY() - antecessorKey.p1.getY())/(antecessorKey.p2.getX() - antecessorKey.p1.getX());
				double m2 = (successorKey.p2.getY() - successorKey.p1.getY())/(successorKey.p2.getX() - successorKey.p1.getX());
				x = (m1*antecessorKey.p1.getX() - m2*successorKey.p1.getX() + successorKey.p1.getY() - antecessorKey.p1.getY())/(m1 - m2);
				y = m2*x - m2*successorKey.p1.getX() + successorKey.p1.getY();
			}
			Point2D pNew = new Point2D.Double(x, y);
			return eventStruct.createEventStructNode(pNew, antecessorNode, successorNode);
		}
	}
}