/*
 * @(#)TwoPolygonsEventStruct.java	1.0 2000/07/22
 *
 */
package rob.math.geom;

import java.util.Comparator;
import java.awt.geom.*;
import rob.util.*;
import rob.exceptions.*;


/**
 * Um objeto <code>TwoPolygonsEventStruct</code> representa uma fila de eventos
 * para o algoritmo de interseces entre dois polgonos. Ela  derivada de um
 * tipo especial de rvore AVL, a <code>AVLTerminatedTree</code>.
 * A estrutura garante remoes, inseres e consultas em tempo O(lg(n)).
 * <p>
 *
 * @version 	1.0 2000/07/22
 * @author 	Roberto Giordano Barra
 * @see         
 */
public class TwoPolygonsEventStruct extends AVLTerminatedTree
{
	// Tipos de eventos que podem existir nessa estrutura.
	public static final int BEGIN = 4;
	public static final int CROSS = 2;
	public static final int END = 1;

	// Guarda a ltima chave retira da estrutura. Sua funo  evitar que ao
	// se criar novos eventos da interseco no se crie eventos anteriores a
	// keyLast, o que poderia gerar loops infinitos.
	protected TwoPolygonsEventStruct.EventStructKey keyLast = null;


	/**
	 * Cria um novo objeto.
	 */
	public TwoPolygonsEventStruct()
	{
		super(new EventStructComparatorProp(), new EventStructResolutor());
	}


	/**
	 * Adiciona um n na estrutura.
	 * @param   thisNode	o n a ser adicionado. 
	 */
	public void insertEvent(AVLTerminatedTree.AVLNode thisNode)
	{
		addNode(thisNode);
	}


	/**
	 * Adiciona um segmento na estrutura. Cria os ns correspondentes e os insere.
	 * @param   p1	um ponto do segmento. 
	 * 		p2	outro ponto do segmento.
	 */
	public void insertSegment(Point2D p1, Point2D p2)
	{
		AVLTerminatedTree.AVLNode nodeInserted = this.createEventStructNode(p1, p2, 0);
		addNode(nodeInserted);
		addNode(this.createEventStructNode(nodeInserted));
	}


	/**
	 * Adiciona um segmento na estrutura. Cria os ns correspondentes e os insere.
	 * @param   p1	um ponto do segmento. 
	 * 		p2	outro ponto do segmento.
	 * 		owner	polgono dono do segmento.
	 */
	public void insertSegment(Point2D p1, Point2D p2, int owner)
	{
		AVLTerminatedTree.AVLNode nodeInserted = this.createEventStructNode(p1, p2, owner);
		addNode(nodeInserted);
		addNode(this.createEventStructNode(nodeInserted));
	}


	/**
	 * Adiciona um segmento na estrutura. Cria os ns correspondentes e os insere.
	 * @param   line	o segmento. 
	 */
	public void insertSegment(Line2D line)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(line.getP1(), line.getP2(), 0);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	/**
	 * Adiciona um segmento na estrutura. Cria os ns correspondentes e os insere.
	 * @param   line	o segmento. 
	 * 		owner	polgono dono do segmento.
	 */
	public void insertSegment(Line2D line, int owner)
	{
		AVLTerminatedTree.AVLNode nodeInseted = this.createEventStructNode(line.getP1(), line.getP2(), owner);
		addNode(nodeInseted);
		addNode(this.createEventStructNode(nodeInseted));
	}


	/**
	 * Retira e devolve o menor elemento da estrutura.
	 * @returns   o evento desejado. 
	 */
	public AVLTerminatedTree.AVLNode getNextAndDelete()
	{
		if (head.left == null || head.left instanceof Terminator)
			return null;
		AVLTerminatedTree.AVLNode thisNode = head.left;
		while (!(thisNode.left instanceof Terminator))
			thisNode = thisNode.left;
		deleteNode(thisNode);
		keyLast = (TwoPolygonsEventStruct.EventStructKey) thisNode.key;
		return thisNode;	
	}


	/**
	 * Cria um n do tipo BEGIN baseado num segmento.
	 * @param   pA	um ponto do segmento. 
	 * 		pB	outro ponto do segmento.
	 * 		owner	polgono dono do segmento.
	 * @returns   o n criado. 
	 */
	public AVLTerminatedTree.AVLNode createEventStructNode(Point2D pA, Point2D pB, int owner)
	{
		return new AVLTerminatedTree.AVLNode(new EventStructKey(pA, pB, owner), null);
	}


	/**
	 * Cria um n do tipo END baseado num n de uma estrutura tipo SweepLine.
	 * @param   toBeRemoved	o n correspondente. 
	 * @returns   o n criado. 
	 */
	public AVLTerminatedTree.AVLNode createEventStructNode(AVLTerminatedTree.AVLNode toBeRemoved)
	{
		return new AVLTerminatedTree.AVLNode(new EventStructKey(((TwoPolygonsEventStruct.EventStructKey) toBeRemoved.key).sweepNode), null);
	}


	/**
	 * Cria um n do tipo CROSS baseado nos ns de uma estrutura tipo SweepLine que
	 * correspondem aos segementos que originaram o ponto de interseco.
	 * @param   pEvt	o ponto de interseco. 
	 * 		lowerOne	um dos ns do segemento de interseco. 
	 * 		upperOne	outro dos ns do segemento de interseco. 
	 * @returns   o n criado. 
	 */
	public AVLTerminatedTree.AVLNode createEventStructNode(Point2D pEvt, AVLRightTerminatedTree.AVLNode lowerOne, AVLRightTerminatedTree.AVLNode upperOne)
	{
		EventStructKey keyNew = new EventCrossKey(pEvt, lowerOne, upperOne);

		// keyNew deve necessariamente ser maior do que keyLast para evitar loops infinitos.
		if (keyLast == null || comp.compare(keyNew, keyLast) > 0)
			return new AVLTerminatedTree.AVLNode(keyNew, null);
		else
			return null;
	}


	/**
	 * Classe que representa um n da estrutura. Pode ser do tipo BEGIN e END. O tipo CROSS tem um
	 * tipo especial descrito abaixo. Cada n equivale a um ponto-evento.
	 */
	public static class EventStructKey
	{
		// Armazena o tipo de n que .
		public int type = 0;

		// Armazena as coordenadas desse ponto-evento.
		public Point2D pEvt = null;

		// Armazena o evento (n) da linha de varredura associado.
		public AVLRightTerminatedTree.AVLNode sweepNode = null;


		/**
		 * Cria um n do tipo BEGIN.
		 * @param   pA	um vrtice do segmento. 
		 * 		pB	outro vrtice do segmento.
		 *		ownerPolygon	identificao do polgono dono da aresta.
		 */
		public EventStructKey(Point2D pA, Point2D pB, int ownerPolygon)
		{
			type = TwoPolygonsEventStruct.BEGIN;
			sweepNode = TwoPolygonsSweepLine.createSweepLineNode(pA, pB, ownerPolygon);
			this.pEvt = ((TwoPolygonsSweepLine.SweepLineKey) sweepNode.key).p1;
		}


		protected EventStructKey()	{}


		/**
		 * Cria um n do tipo END.
		 * @param   toBeRemoved	evento da linha de varredura associado. 
		 */
		public EventStructKey(AVLRightTerminatedTree.AVLNode toBeRemoved)
		{
			type = TwoPolygonsEventStruct.END;
			sweepNode = toBeRemoved;
			this.pEvt = ((TwoPolygonsSweepLine.SweepLineKey) sweepNode.key).p2;
		}


		public String toString()
		{
			if (type == TwoPolygonsEventStruct.BEGIN)
				return " BEGIN:(" + pEvt.getX() + ", " + pEvt.getY() + ")";
			else if (type == TwoPolygonsEventStruct.CROSS)
				return " CROSS:(" + pEvt.getX() + ", " + pEvt.getY() + ")";
			else if (type == TwoPolygonsEventStruct.END)
				return " END:(" + pEvt.getX() + ", " + pEvt.getY() + ")";
			else
				return " fuck:(" + pEvt.getX() + ", " + pEvt.getY() + ")";
		}
	}


	/**
	 * Classe que representa um ponto-evento (n) do tipo CROSS da estrutura. 
	 */
	public static class EventCrossKey extends EventStructKey
	{
		// Armazena o evento da linha de varredura associado ao segmento que  responsvel
		// pela inerseco e que se localiza acima do outro segemento responsvel, para x menor
		// do que a x-coordenada da interseco.
		public AVLRightTerminatedTree.AVLNode successorNode = null;

		// Armazena o evento da linha de varredura associado ao segmento que  responsvel
		// pela inerseco e que se localiza abaixo do outro segemento responsvel, para x menor
		// do que a x-coordenada da interseco.
		public AVLRightTerminatedTree.AVLNode antecessorNode = null;


		/**
		 * Cria um n.
		 * @param   pEvt	as coordenadas da interseco. 
		 * 		lowerOne	segmento responsvel superior para x-coordenadas anteriores a pEvt.
		 *		upperOne	segmento responsvel inferior para x-coordenadas anteriores a pEvt.
		 */
		public EventCrossKey(Point2D pEvt, AVLRightTerminatedTree.AVLNode lowerOne, AVLRightTerminatedTree.AVLNode upperOne)
		{
			super();
			type = TwoPolygonsEventStruct.CROSS;
			this.pEvt = pEvt;
			successorNode = upperOne;
			antecessorNode = lowerOne;
		}
	}


	/**
	 * Classe que determina se um n da estrututura TwoPoligonSweepLine  maior, menor ou
	 * igual a outro n.
	 */
	public static class EventStructComparatorProp implements Comparator
	{
		/**
		 * Mtodo que realiza a comparao entre o1 e o2, onde o1 e o2 so ns da estrutura em questo.
		 * O critrio  o seguinte: o1  maior que o2 se:
		 *	1- O pEvt da chave de  o1  maior com relao a x-coordenada do que pEvt da chave de  o2.
		 *	2- O pEvt da chave de  o1  maior com relao a y-coordenada do que pEvt da chave de  o2.
		 *	3- Se o1  do tipo BEGIN e o2 no.
		 *	4- Se o1  do tipo CROSS e o2  do tipo END.
		 *	(Se passar de 4, os tipos so iguais.)
		 *	(a- tipo == CROSS; b- tipo != CROSS)
		 *	5a- Dos antecessores: se o ponto p2 de o2 estiver a direita do segmento  p1->p2 de o1.
		 *	6a- Dos sucessores:   se o ponto p2 de o2 estiver a direita do segmento  p1->p2 de o1.
		 *	7a- So iguais!
		 *	5b- Se o ponto p2 de o1 estiver a direita do segmento  p1->p2 de o2.
		 *	6b- Se a coordenada x de p2 de o1 for maior do que a coordenada x de p2 de o2.
		 * @param   o1	uma das chaves a ser comparadas. 
		 * 		o2	outra das chaves a ser comparadas. 
		 */
		public int compare(Object o1, Object o2)
		{
			EventStructKey key1 = (EventStructKey) o1;
			EventStructKey key2 = (EventStructKey) o2;
			
			int result = GeomOperations.compareByX(key1.pEvt, key2.pEvt);
			if (result != 0)
				return result;
			else
			{
				if (key1.type != key2.type)
					return key1.type - key2.type;
				else
				{
					double area2 = 0;
					if (key1.type == TwoPolygonsEventStruct.CROSS)
					{
						// Verifica qual dos segmentos antecessores dos eventos
						// sendo comparados est embaixo. Se o atual estiver mais
						// embaixo, para.
						EventCrossKey crossKey1 = (EventCrossKey) o1;
						EventCrossKey crossKey2 = (EventCrossKey) o2;
						area2 = GeomOperations.area2(((TwoPolygonsSweepLine.SweepLineKey)crossKey1.antecessorNode.key).p1, 
											((TwoPolygonsSweepLine.SweepLineKey)crossKey1.antecessorNode.key).p2, 
											((TwoPolygonsSweepLine.SweepLineKey)crossKey2.antecessorNode.key).p2);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							// Verifica se qual dos segmentos sucessores dos eventos
							// sendo comparados est embaixo. Se o atual estiver mais
							// embaixo, para.
							area2 = GeomOperations.area2(((TwoPolygonsSweepLine.SweepLineKey)crossKey1.successorNode.key).p1, 
												((TwoPolygonsSweepLine.SweepLineKey)crossKey1.successorNode.key).p2, 
												((TwoPolygonsSweepLine.SweepLineKey)crossKey2.successorNode.key).p2);
							if (area2 != 0)
								return area2 > 0 ? 1 : -1;
							else
								return 0;
						}
					}
					else
					{
						if (key1.type == TwoPolygonsEventStruct.BEGIN)
							area2 = GeomOperations.area2(((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p1, 
												((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p2, 
												((TwoPolygonsSweepLine.SweepLineKey)key1.sweepNode.key).p2);
						else
							area2 = GeomOperations.area2(((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p1, 
												((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p2, 
												((TwoPolygonsSweepLine.SweepLineKey)key1.sweepNode.key).p1);
						if (area2 != 0)
							return area2 > 0 ? 1 : -1;
						else
						{
							if (((TwoPolygonsSweepLine.SweepLineKey)key1.sweepNode.key).p2.getX() > ((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p2.getX())
								return 1;
							else if (((TwoPolygonsSweepLine.SweepLineKey)key1.sweepNode.key).p2.getX() < ((TwoPolygonsSweepLine.SweepLineKey)key2.sweepNode.key).p2.getX())
								return -1;
							else
								return 0;
						}
					}
				}
			}
		}
	}


	/**
	 * Classe que determina o que fazer caso dois forem dados como iguais pelo comparator.
	 * Nesse caso no faz nada e mantm a deciso do comparator.
	 */
	public static class EventStructResolutor implements Resolutor
	{		
		public int resolve(INodeMap nodeOne, INodeMap nodeTwo)	{return 0;}
		public int resolveKeys(Object keyOne, Object keyTwo)	{return 0;}
	}
}