/*
 * @(#)TwoPolygonsSweepLine.java	1.0 2000/07/22
 *
 */
package rob.math.geom;

import rob.util.*;
import java.awt.geom.*;
import java.util.Comparator;


/**
 * Um objeto <code>TwoPolygonsSweepLine</code> representa uma linha de varredura
 * para o algoritmo de interseces netre dois polgonos. Ela  derivada de um
 * tipo especial de rvore AVL, a <code>AVLRightTerminatedTree</code>.
 * A estrutura garante remoes, inseres e consultas em tempo O(lg(n)).
 * <p>
 *
 * @version 	1.0 2000/07/22
 * @author 	Roberto Giordano Barra
 * @see         
 */
public class TwoPolygonsSweepLine extends AVLRightTerminatedTree
{
	/**
	 * Cria um novo objeto que procura que tem seus terminadores (terminators) criados
	 * pelo objeto passado.
	 * @param   termFactory	o objeto que  encarregado de produzir terminadores para a 
	 * linha de varredura criada.
	 */
	public TwoPolygonsSweepLine(TerminatorFactory termFactory)
	{
		super(new SweepLineComparator(), new SweepLineResolutor(), termFactory);
	}


	/**
	 * Adiciona um n na estrutura.
	 * @param   thisNode	o n a ser adicionado. 
	 */
	public void insert(AVLRightTerminatedTree.AVLNode thisNode)
	{
		addNode(thisNode);
	}


	/**
	 * Remove um n da estrutura.
	 * @param   thisNode	o n a ser removido. 
	 */
	public void delete(AVLRightTerminatedTree.AVLNode thisNode)
	{	
		deleteNode(thisNode);
	}


	/**
	 * Cria um n para a estrutura baseado no segmento passado seguindo o padro adotado 
	 * para um n. Veja detalhes em SweepNodeKey e SweepNodeElement.
	 * @param   pA	um vrtice do segmento. 
	 * 		pB	outro vrtice do segmento.
	 * 		ownerPolygon	identificao do polgono dono do segmento passado.
	 */
	public static AVLRightTerminatedTree.AVLNode createSweepLineNode(Point2D pA, Point2D pB, int ownerPolygon)
	{
		return new AVLRightTerminatedTree.AVLNode(new SweepLineKey(pA, pB), new SweepLineElement(ownerPolygon));
	}


	/**
	 * Classe que representa a chave de um n da estrutura TwoPolygonSweepLine.  na chave
	 * que a informao do segmento em s est contida, pois  necessria para ordenar o n
	 * na rvore.
	 */
	public static class SweepLineKey
	{
		// p2 ser sempre maior ou igual a p1 com relao a:
		// 	1- x-coordenada.
		// 	2- y-coordenada.
		public Point2D p1 = null, p2 = null;


		/**
		 * Cria um n.
		 * @param   pA	um vrtice do segmento. 
		 * 		pB	outro vrtice do segmento.
		 */
		public SweepLineKey(Point2D pA, Point2D pB)
		{
			if (GeomOperations.compareByX(pA, pB) >= 0)	{p1 = pB; p2 = pA;}
			else								{p1 = pA; p2 = pB;}
		}


		public String toString()
		{
			return "[p1:(" + p1.getX() + ", " + p1.getY() + ")p2:(" + p2.getX() + ", " + p2.getY() + ")]";
		}
	}


	/**
	 * Classe que representa o elemento de um n da estrutura TwoPolygonSweepLine.  no elemento
	 * que a informao sobre o dono da aresta  armazenada.  tambm guardada informao para 
	 * ser utilizada na parte de gerao da lista de sada do algoritmo.
	 */
	public static class SweepLineElement
	{
		// Armazena a identificao do polgno dono da aresta a que o elemento est
		// associado.
		private int owner;

		// Armazena o vrtice na estrutura que armazena os vrtices de sada do algoritmo
		// que est associado a este elemento.
		private TwoPolygonsOutputGenerator.Vertex vertex = null;

		// Armazena a que tipo de cadeia de vrtices (na estrutura que armazena os vrtices de sada 
		// do algoritmo) a aresta associada a esse elemento se encontra. Mais informaes na classe
		// TwoPolygonOutputGenerator.
		private boolean upperChain = false;

		/**
		 * Cria um elemento.
		 * @param   ownerNew	a identificao do polgono a esse associado. 
		 */
		public SweepLineElement(int ownerNew)	{owner = ownerNew;}
	
	
		public int getOwner()	{return owner;}
		public void setOwner(int newOwner)	{owner = newOwner;}


		public TwoPolygonsOutputGenerator.Vertex getVertex()	{return vertex;}
		public void setVertex(TwoPolygonsOutputGenerator.Vertex newVertex)	{vertex = newVertex;}


		public boolean getUpperChain()	{return upperChain;}
		public void setUpperChain(boolean newUpperChain)	{upperChain = newUpperChain;}
	}


	/**
	 * Classe que determina se um n da estrututura TwoPoligonSweepLine  maior, menor ou
	 * igual a outro n.
	 */
	public static class SweepLineComparator implements Comparator
	{
	/**
	 * Mtodo que realiza a comparao entre o1 e o2, onde o1 e o2 so ns da estrutura em questo.
	 * O critrio  o seguinte: o1  maior que o2 se:
	 *	1- O ponto p1 da chave de  o1 est a direita do segmento orientado p1->p2 de o2.
	 *	2- O ponto p2 da chave de  o1 est a direita do segmento orientado p1->p2 de o2.
	 *	3- Se o ponto p2 de o1 for maior ou igual em x ao ponto p2 de o2.
	 * @param   ownerNew	a identificao do polgono a esse associado. 
	 */
		public int compare(Object o1, Object o2)
		{
			SweepLineKey key1 = (SweepLineKey) o1;
			SweepLineKey key2 = (SweepLineKey) o2;

			double area2 = GeomOperations.area2(key2.p1, key2.p2, key1.p1);
			if (area2 != 0)
				return area2 > 0 ? 1 : -1;
			else
			{
				area2 = GeomOperations.area2(key2.p1, key2.p2, key1.p2);
				if (area2 != 0)
					return area2 > 0 ? 1 : -1;
				else
					return key1.p2.getX() > key2.p2.getX() ? 1 : -1;
			}				
		}
	}


	/**
	 * Classe que determina o que fazer caso dois forem dados como iguais pelo comparator.
	 * Nesse caso no faz nada e mantm a deciso do comparator.
	 */
	public static class SweepLineResolutor implements Resolutor
	{		
		public int resolve(INodeMap nodeOne, INodeMap nodeTwo)	{return 0;}
		public int resolveKeys(Object keyOne, Object keyTwo)	{return 0;}
	}





	/**
	 * Mtodo para debug. Imprime o valor dos terminators da estrutura.
	 */
	public void debugTerminators()
	{
		if (head.left == null || head.left instanceof AVLRightTerminatedTree.RightTerminator)
			return;
		AVLRightTerminatedTree.AVLNode thisNode = head.left;
		while (!(thisNode instanceof AVLRightTerminatedTree.RightTerminator))
			thisNode = thisNode.left;

		AVLRightTerminatedTree.RightTerminator thisTerm = (AVLRightTerminatedTree.RightTerminator) thisNode;
		while (thisNode != null)
		{
			if (thisNode.parent != null)
				System.out.print(" | parent:" + thisNode.parent.key.toString() + ", " + ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) ((AVLRightTerminatedTree.RightTerminator) thisNode).element).getOwner());
			else
				System.out.print(" | parent:" + thisNode.parent);
			thisNode = thisNode.right;
		}
		System.out.println("/n\n");

		while (thisTerm != null)
		{
			System.out.print(" | parent:" + thisTerm.parent.key.toString() + ", " + ((TwoPolygonsRegionsMarkersTerminatorFactory.Element) thisTerm.element).getOwner());
			thisTerm = (AVLRightTerminatedTree.RightTerminator) thisTerm.next();
		}
	}
}