package rob.util;
import java.util.*;


public class RobTreeMap extends AbstractMap implements SortedMap, Cloneable, java.io.Serializable
{
	protected Comparator comparator = null;
	protected transient Entry root = null;
	protected transient int size = 0;
	protected transient int modCount = 0;

	// Views
	private transient Set		keySet = null;
	private transient Set		entrySet = null;
	private transient Collection	values = null;

	// Types of Iterators
	private static final int KEYS = 0;
	private static final int VALUES = 1;
	private static final int ENTRIES = 2;

	protected static final boolean RED   = false;
	protected static final boolean BLACK = true;

	public RobTreeMap()	{}


	public RobTreeMap(Comparator c)	{this.comparator = c;}
	public RobTreeMap(Map m)	{putAll(m);}


	public RobTreeMap(SortedMap m)
	{
		comparator = m.comparator();
		try 
		{
			buildFromSorted(m.size(), m.entrySet().iterator(), null, null);	
		} 
		catch (java.io.IOException cannotHappen) {} 
		catch (ClassNotFoundException cannotHappen) {}
    }



	protected void incrementSize()   { modCount++; size++; }
	protected void decrementSize()   { modCount++; size--; }


    // Query Operations


	public Comparator comparator()	{return comparator;}
	public Object firstKey()	{return key(firstEntry());}
	public int size()	{return size;}
	public Object lastKey()	{return key(lastEntry());}
	public boolean containsKey(Object key)	{return getEntry(key) != null;}


	public boolean containsValue(Object value)
	{
		return (root == null ? false : (value == null ? valueSearchNull(root) : valueSearchNonNull(root, value)));
	}


	protected boolean valueSearchNull(Entry n)
	{
      	if (n.value == null)
	            return true;
		return (n.left  != null && valueSearchNull(n.left)) || (n.right != null && valueSearchNull(n.right));
	}


	protected boolean valueSearchNonNull(Entry n, Object value)
	{
		if (value.equals(n.value))
            	return true;
		return (n.left  != null && valueSearchNonNull(n.left, value)) || (n.right != null && valueSearchNonNull(n.right, value));
	}


	public Object get(Object key)
	{
		Entry p = getEntry(key);
		return (p == null ? null : p.value);
	}


	public void putAll(Map map)
	{
		int mapSize = map.size();
		if (size == 0 && mapSize != 0 && map instanceof SortedMap)
		{
            	Comparator c = ((SortedMap)map).comparator();
	            if (c == comparator || (c != null && c.equals(comparator)))
			{
      	      	++modCount;
				try
				{
					buildFromSorted(mapSize, map.entrySet().iterator(), null, null);
				}
				catch (java.io.IOException cannotHappen) {} 
				catch (ClassNotFoundException cannotHappen) {}
				return;
			}
		}
		super.putAll(map);
	}


	protected Entry getEntry(Object key)
	{
		Entry p = root;
		while (p != null)	
		{
			int cmp = compare(key,p.key);
			if (cmp == 0)
				return p;
			else if (cmp < 0)
				p = p.left;
			else
				p = p.right;
		}
		return null;
	}


/**
 *	Retorna o objeto associado com a key passada ou se tal key no existe retorna
 * o objeto do conjunto de objetos com key maior do que a passada o com key mais prxima
 * a key passada. Esse elemento  procurado quando se chega numa das folhas da rvore e se
 * descobre que no h key na rvore. Da,  s ir voltando na arvore at achar um n cujo 
 * o pai tem ele como filho esquerdo.
 *	Se no tiver tal n, isso significa qoe todos os ns da vore so menores do que key,
 * e  retornado null.
 */
	protected Entry getCeilEntry(Object key)
	{
		Entry p = root;
		if (p == null)
			return null;
		while (true)
		{
			int cmp = compare(key, p.key);
			if (cmp == 0)
				return p;
			else if (cmp < 0)
			{
				if (p.left != null)
					p = p.left;
				else
					return p;
			}
			else
			{
				if (p.right != null)
				    p = p.right;
				else
				{
					Entry parent = p.parent;
					Entry ch = p;
					while (parent != null && ch == parent.right)
					{
						ch = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
	}



/**
 *	Retorna o objeto do conjunto de objetos com key menor do que a passada o com key mais 
 * prxima a key passada. Esse elemento  procurado quando se chega numa das folhas da rvore 
 * e se descobre que no h key na rvore. Da,  s ir voltando na arvore at achar um n 
 * cujo o pai tem ele como filho direito.
 *	Se no tiver tal n, isso significa qoe todos os ns da vore so menores do que key,
 * e  retornado null.
 */

	protected Entry getPreviousEntry(Entry entry)
	{
		if (entry.left == null)
		{
			Entry pai = entry.parent;
			while (entry.parent != null && entry.parent.right != entry)
				entry = entry.parent;
			return entry.parent;
		}
		else
		{
			entry = entry.left;
			while (entry.right != null)
				entry = entry.right;
			return entry;
		}
	}	


	protected Entry getPrecedingEntry(Object key)
	{
		Entry p = root;
		if (p == null)
			return null;
		while (true)
		{
			int cmp = compare(key, p.key);
	            if (cmp > 0)
			{
				if (p.right != null)
				    p = p.right;
				else
			    		return p;
		    	} 
			else	
			{
				if (p.left != null)
					p = p.left;
				else	
				{
					Entry parent = p.parent;
					Entry ch = p;
					while (parent != null && ch == parent.left)
					{
						ch = parent;
						parent = parent.parent;
				    	}
					return parent;
				}
			}
		}	
	}


	public Object getPrecedingKey(Object key)
	{
 		Object retValue = getPrecedingEntry(key);
		if (retValue != null)
			retValue = ((Entry) retValue).key;
		return retValue;
	}

/**
 *	Retorna o objeto do conjunto de objetos com key menor do que a passada o com key mais 
 * prxima a key passada. Esse elemento  procurado quando se chega numa das folhas da rvore 
 * e se descobre que no h key na rvore. Da,  s ir voltando na arvore at achar um n 
 * cujo o pai tem ele como filho direito.
 *	Se no tiver tal n, isso significa qoe todos os ns da vore so menores do que key,
 * e  retornado null.
 */


	protected Entry getNextEntry(Entry entry)
	{
		if (entry.right == null)
		{
			Entry pai = entry.parent;
			while (entry.parent != null && entry.parent.left != entry)
				entry = entry.parent;
			return entry.parent;
		}
		else
		{
			entry = entry.right;
			while (entry.left != null)
				entry = entry.left;
			return entry;
		}
	}	


	protected Entry getNextEntry(Object key)
	{
		Entry p = root;
		if (p == null)
			return null;
		while (true)
		{
			int cmp = compare(key, p.key);
	            if (cmp <= 0)
			{
				if (p.left != null)
				    p = p.left;
				else
			    		return p;
		    	} 
			else	
			{
				if (p.right != null)
					p = p.right;
				else	
				{
					Entry parent = p.parent;
					Entry ch = p;
					while (parent != null && ch == parent.right)
					{
						ch = parent;
						parent = parent.parent;
				    	}
					return parent;
				}
			}
		}	
	}


	protected Entry getFirstHigherEntry(Object key)
	{
		Entry p = root;
		if (p == null)
			return null;
		while (true)
		{
			int cmp = compare(key, p.key);
	            if (cmp <= 0)
			{
				if (p.left != null)
				    p = p.left;
				else
			    		return p;
		    	} 
			else	
			{
				return getNextEntry(p);
			}
		}	
	}


	public Object getNextKey(Object key)
	{
		Object retValue = getNextEntry(key);
		if (retValue != null)
			retValue = ((Entry) retValue).key;
		return retValue;
	}



//	S usar quando  garantido a presena dos elementos.
	protected void swapWithPreviousEntry(Entry entry)
	{
		Entry prevP = null;
		if (entry.left == null)
		{
			prevP = entry.parent;
			Entry ch = entry;
			while (prevP != null && ch == prevP.left)
			{
				ch = prevP;
				prevP = prevP.parent;
		    	}
		}
		else
		{
			prevP = entry.left;
			while (prevP.right != null)
				prevP = prevP.right;
		}
		swapPosition(entry, prevP);
	}


//	S usar quando  garantido a presena dos elementos.
	protected void swapWithNextEntry(Entry entry)
	{
		Entry nextP = null;
		if (entry == null)
			System.out.print("BUNDO! -> entry == null ");
		if (entry.right == null)
		{
			nextP = entry.parent;
			Entry ch = entry;
			while (nextP != null && ch == nextP.right)
			{
				ch = nextP;
				nextP = nextP.parent;
		    	}
			if (nextP == null)
				System.out.print("BUNDO! -> nextP == null ");
		}
		else
		{
			nextP = entry.right;
			if (nextP == null)
				System.out.print("BUNDO2! -> nextP == null ");
			while (nextP.left != null)
				nextP = nextP.left;
		}
		swapPosition(nextP, entry);
	}


	protected void swapNextPrev(Entry entry, Entry prevP)
	{
		if (entry == prevP)
			System.out.print("MERRRRRDA!-> entry == prevP ");
		Entry tmp;
		tmp = entry.left;     entry.left   = prevP.left;     prevP.left   = tmp;
		tmp = entry.right;    entry.right  = prevP.right;    prevP.right  = tmp;
		tmp = entry.parent;   entry.parent = prevP.parent;   prevP.parent = tmp;
		if (entry.parent == entry)	// entry agora  filho de prevP
		{
			System.out.print(" entry.parent == entry ");
			entry.parent = prevP;
			if (entry.left != null)		entry.left.parent  = entry;
			if (entry.right != null)	entry.right.parent = entry;

			if (prevP.left == prevP)
			{
				prevP.left = entry;
				if (prevP.right != null)	prevP.right.parent = prevP;
			}
			else
			{
				prevP.right = entry;
				if (prevP.left != null)		prevP.left.parent  = prevP;
			}
			if (prevP.parent != null)
			{
				if (prevP.parent.left == entry)	prevP.parent.left  = prevP;
				else						prevP.parent.right = prevP;
			}
			else		root = prevP;
		}
		else if (prevP.parent == prevP)	// prevP agora  filho entry.
		{
			System.out.print(" prevP.parent == prevP ");
			if (prevP.left == null && prevP.right == null)
				System.out.print(" 1 ");
			if (entry.left == null && entry.parent == null)
				System.out.print(" 2 ");

			prevP.parent = entry;
			if (prevP.left != null)		prevP.left.parent  = prevP;
			if (prevP.right != null)	prevP.right.parent = prevP;

			if (entry.left == entry)
			{
				entry.left = prevP;
				if (entry.right != null)	entry.right.parent = entry;
			}
			else
			{
				entry.right = prevP;
				if (entry.left != null)		entry.left.parent  = entry;
			}
			if (entry.parent != null)
			{
				if (entry.parent.left == prevP)	entry.parent.left  = entry;
				else						entry.parent.right = entry;
			}
			else		root = entry;
		}
		else
		{
			if (entry.left != null)		entry.left.parent  = entry;
			if (prevP.left != null)		prevP.left.parent  = prevP;
			if (entry.right != null)	entry.right.parent = entry;
			if (prevP.right != null)	prevP.right.parent = prevP;
			if (prevP.parent != null)
			{
				if (prevP.parent.left == entry)	prevP.parent.left  = prevP;
				else						prevP.parent.right = prevP;
			}
			if (entry.parent != null)
			{
				if (entry.parent.left == prevP)	entry.parent.left  = entry;
				else						entry.parent.right = entry;
			}
		}
		boolean bTmp = entry.color; entry.color = prevP.color; prevP.color = bTmp;
		if (prevP.left == null && prevP.right == null)
			System.out.print(" 1 ");
		if (entry.left == null && entry.parent == null)
			System.out.print(" 2 ");
	}


	private static Object key(Entry e)
	{
		if (e ==null)
			throw new NoSuchElementException();
		return e.key;
	}


	public Object put(Object key, Object value)
	{
		Entry t = root;

		if (t == null)
		{
			incrementSize();
			root = new Entry(key, value, null);
			return null;
		}
		while (true)
		{
			int cmp = compare(key, t.key);
			if (cmp <= 0)
			{
				if (t.left != null)
				    t = t.left;
				else
				{
					incrementSize();
		    			t.left = new Entry(key, value, t);
					fixAfterInsertion(t.left);
					return null;
				}
	    		}
			else	
			{
				if (t.right != null)
					t = t.right;
				else
				{
		    			incrementSize();
					t.right = new Entry(key, value, t);
					fixAfterInsertion(t.right);
					return null;
				}
			}
		}
	}


	public Object remove(Object key)
	{
		Entry p = getEntry(key);
		if (p == null)
			return null;
		Object oldValue = p.value;
		deleteEntry(p);
		return oldValue;
    	}


	public void clear()
	{
		modCount++;
		size = 0;
		root = null;
	}


	public Object clone()	{return new RobTreeMap(this);}


	public Set keySet()
	{
		if (keySet == null)
		{
			keySet = new AbstractSet()
			{
				public java.util.Iterator iterator() {return new Iterator(KEYS);}
				public int size() {return RobTreeMap.this.size();}
				public boolean contains(Object o) {return containsKey(o);}
				public boolean remove(Object o) {return RobTreeMap.this.remove(o) != null;}
				public void clear() {RobTreeMap.this.clear();}
			};
		}
		return keySet;	
	}


	public Collection values()
	{
		if (values == null) 
		{
      	      values = new AbstractCollection()
			{
				public java.util.Iterator iterator() {return new Iterator(VALUES);}
               		public int size() {return RobTreeMap.this.size();}


				public boolean contains(Object o)
				{
                    		for (Entry e = firstEntry(); e != null; e = successor(e))
						if (valEquals(e.getValue(), o))
							return true;
					return false;
				}


				public boolean remove(Object o)
				{
            			for (Entry e = firstEntry(); e != null; e = successor(e))
						if (valEquals(e.getValue(), o))
						{
                            			deleteEntry(e);
							return true;
						}
					return false;
				}


				public void clear() {RobTreeMap.this.clear();}
			};
		}
        	return values;
	}


	public Set entrySet()
	{
		if (entrySet == null)
		{
			entrySet = new AbstractSet()
			{
				public java.util.Iterator iterator() {return new Iterator(ENTRIES);}


				public boolean contains(Object o)
				{
                  		if (!(o instanceof Map.Entry))
                    			return false;
                    		Map.Entry entry = (Map.Entry)o;
	                    	Object value = entry.getValue();
                    		Entry p = getEntry(entry.getKey());
                    		return p != null && valEquals(p.getValue(), value);
				}


				public boolean remove(Object o)
				{
                  		if (!(o instanceof Map.Entry))
                        		return false;
	                    	Map.Entry entry = (Map.Entry)o;
                    		Object value = entry.getValue();
                    		Entry p = getEntry(entry.getKey());
                    		if (p != null && valEquals(p.getValue(), value))
					{
                        		deleteEntry(p);
                        		return true;
					}
					return false;
				}


				public int size() {return RobTreeMap.this.size();}	
				public void clear() {RobTreeMap.this.clear();}
			};
     		}
		return entrySet;
	}


	public SortedMap subMap(Object fromKey, Object toKey)		{return new SubMap(fromKey, toKey);}
	public SortedMap headMap(Object toKey)		{return new SubMap(toKey, true);}
	public SortedMap tailMap(Object fromKey) 		{return new SubMap(fromKey, false);}


	private class SubMap extends AbstractMap implements SortedMap, java.io.Serializable
	{
      	private static final long serialVersionUID = -6520786458950516097L;
		private boolean fromStart = false, toEnd = false;
		private Object  fromKey,           toKey;
		private transient Set entrySet = new EntrySetView();


		SubMap(Object fromKey, Object toKey)
		{
			if (compare(fromKey, toKey) > 0)
				throw new IllegalArgumentException("fromKey > toKey");
			this.fromKey = fromKey;
	    		this.toKey = toKey;
		}


		SubMap(Object key, boolean headMap)
		{
			if (headMap) 
			{
				fromStart = true;
				toKey = key;
			} 
			else
			{
				toEnd = true;
                		fromKey = key;
            	}
		}


		SubMap(boolean fromStart, Object fromKey, boolean toEnd, Object toKey)
		{
            	this.fromStart = fromStart;
	            this.fromKey= fromKey;
      	      this.toEnd = toEnd;
            	this.toKey = toKey;
		}


		public boolean isEmpty() {return entrySet.isEmpty();}
		public boolean containsKey(Object key) {return inRange(key) && RobTreeMap.this.containsKey(key);}
		public Set entrySet() {return entrySet;}
		public Comparator comparator()	{return comparator;}
		public Object firstKey()	{return key(fromStart ? firstEntry() : getCeilEntry(fromKey));}
		public Object lastKey()	{return key(toEnd ? lastEntry() : getPrecedingEntry(toKey));}
	

		public Object get(Object key)
		{
			if (!inRange(key))
				return null;
			return RobTreeMap.this.get(key);
		}


		public Object put(Object key, Object value)
		{
			if (!inRange(key))
				throw new IllegalArgumentException("key out of range");
			return RobTreeMap.this.put(key, value);
		}


		private class EntrySetView extends AbstractSet
		{
			private transient int size = -1, sizeModCount;
			public int size()
			{
				if (size == -1 || sizeModCount != RobTreeMap.this.modCount)
				{
					size = 0;  sizeModCount = RobTreeMap.this.modCount;
					java.util.Iterator i = iterator();
					while (i.hasNext())
					{
						size++;
						i.next();
					}
			}
			return size;
		}

		public boolean isEmpty()	{return !iterator().hasNext();}

		public boolean contains(Object o)
		{
			if (!(o instanceof Map.Entry))
				return false;
			Map.Entry entry = (Map.Entry) o;
			Object key = entry.getKey();
			if (!inRange(key))
				return false;
			RobTreeMap.Entry node = getEntry(key);
			return node != null && valEquals(node.getValue(), entry.getValue());
		}


		public boolean remove(Object o)
		{
			if (!(o instanceof Map.Entry))
		    		return false;
			Map.Entry entry = (Map.Entry)o;
			Object key = entry.getKey();
			if (!inRange(key))
				return false;
			RobTreeMap.Entry node = getEntry(key);
			if (node!=null && valEquals(node.getValue(),entry.getValue()))
			{
				deleteEntry(node);
				return true;
			}
			return false;
		}


		public java.util.Iterator iterator()	{return new Iterator((fromStart ? firstEntry() : getCeilEntry(fromKey)), (toEnd ? null : getCeilEntry(toKey)));}
	}



		public SortedMap subMap(Object fromKey, Object toKey)
		{
            	if (!inRange(fromKey))
				throw new IllegalArgumentException("fromKey out of range");
	            if (!inRange2(toKey))
      			throw new IllegalArgumentException("toKey out of range");
			return new SubMap(fromKey, toKey);
		}


		public SortedMap headMap(Object toKey)
		{
            	if (!inRange2(toKey))
				throw new IllegalArgumentException("toKey out of range");
			return new SubMap(fromStart, fromKey, false, toKey);
		}


		public SortedMap tailMap(Object fromKey)
		{
            	if (!inRange(fromKey))	
				throw new IllegalArgumentException("fromKey out of range");
            	return new SubMap(false, fromKey, toEnd, toKey);
		}


		private boolean inRange(Object key)
		{
	    		return (fromStart || compare(key, fromKey) >= 0) && (toEnd     || compare(key, toKey)   <  0);
		}


		// This form allows the high endpoint (as well as all legit keys)
		private boolean inRange2(Object key)
		{
			return (fromStart || compare(key, fromKey) >= 0) && (toEnd     || compare(key, toKey)   <= 0);
		}
	}


    /**
     * RobTreeMap Iterator.
     */
	private class Iterator implements java.util.Iterator
	{
		private int type;
		private int expectedModCount = RobTreeMap.this.modCount;
		private Entry lastReturned = null;
		private Entry next;
		private Entry firstExcluded = null;


		Iterator(int type)
		{
			this.type = type;
			next = firstEntry();
		}


		Iterator(Entry first, Entry firstExcluded)
		{
			type = ENTRIES;
			next = first;
			this.firstExcluded = firstExcluded;
		}


		public boolean hasNext() {return next != firstExcluded;}


		public Object next()
		{
			if (next == firstExcluded)
				throw new NoSuchElementException();
			if (modCount != expectedModCount)
				throw new ConcurrentModificationException();
			lastReturned = next;
			next = successor(next);
			return (type == KEYS ? lastReturned.key : (type == VALUES ? lastReturned.value : lastReturned));
		}


		public void remove()
		{
			if (lastReturned == null)
				throw new IllegalStateException();
			if (modCount != expectedModCount)
				throw new ConcurrentModificationException();
			deleteEntry(lastReturned);
			expectedModCount++;
			lastReturned = null;
		}
	}


	protected int compare(Object k1, Object k2)
	{
		return (comparator==null ? ((Comparable)k1).compareTo(k2): comparator.compare(k1, k2));
	}


	protected static boolean valEquals(Object o1, Object o2)
	{
		return (o1==null ? o2==null : o1.equals(o2));
	}


    /**
     * Node in the Tree.  Doubles as a means to pass key-value pairs back to
     * user (see Map.Entry).
     */
	public static class Entry implements Map.Entry
	{
		public Object key;
		public Object value;
		public Entry left = null;
		public Entry right = null;
		public Entry parent;
		public boolean color = BLACK;

		public Entry(Object key, Object value, Entry parent)
		{ 
			this.key = key;
			this.value = value;
			this.parent = parent;
		}

		public Object getKey()	{return key;}
		public Object getValue()	{return value;}


		public Object setValue(Object value)
		{
			Object oldValue = this.value;
			this.value = value;
			return oldValue;
		}


		public boolean equals(Object o)
		{
			if (!(o instanceof Map.Entry))
				return false;
			Map.Entry e = (Map.Entry)o;
			return valEquals(key,e.getKey()) && valEquals(value,e.getValue());
		}


		public int hashCode()
		{
			int keyHash = (key==null ? 0 : key.hashCode());
			int valueHash = (value==null ? 0 : value.hashCode());
			return keyHash ^ valueHash;
		}


		public String toString()	{return key + "=" + value;}
	}


	protected Entry firstEntry()
	{
		Entry p = root;
		if (p != null)
	    		while (p.left != null)
				p = p.left;
		return p;
	}

	
	protected Entry lastEntry()
	{
		Entry p = root;
		if (p != null)
			while (p.right != null)
				p = p.right;
		return p;
	}


	protected Entry successor(Entry t)
	{
		if (t == null)
			return null;
		else if (t.right != null)
		{
			Entry p = t.right;
			while (p.left != null)
				p = p.left;
			return p;
		}
		else
		{
			Entry p = t.parent;
			Entry ch = t;
			while (p != null && ch == p.right)
			{
				ch = p;
				p = p.parent;
			}
			return p;
		}
	}


    /**
     * Balancing operations.
     *
     * Implementations of rebalancings during insertion and deletion are
     * slightly different than the CLR version.  Rather than using dummy
     * nilnodes, we use a set of accessors that deal properly with null.  They
     * are used to avoid messiness surrounding nullness checks in the main
     * algorithms.
     */
	private static boolean colorOf(Entry p)	{return (p == null ? BLACK : p.color);}
	private static Entry  parentOf(Entry p)	{return (p == null ? null: p.parent);}
	private static void setColor(Entry p, boolean c) {if (p != null)  p.color = c;}
	private static Entry  leftOf(Entry p)	 {return (p == null)? null: p.left;}
	private static Entry  rightOf(Entry p)	{return (p == null)? null: p.right;}


	/** From CLR **/
	protected void rotateLeft(Entry p)
	{
		Entry r = p.right;
		p.right = r.left;
		if (r.left != null)
			r.left.parent = p;
		r.parent = p.parent;
		if (p.parent == null)
			root = r;
		else if (p.parent.left == p)
			p.parent.left = r;
		else
			p.parent.right = r;
		r.left = p;
		p.parent = r;
	}


	/** From CLR **/
	protected void rotateRight(Entry p)
	{
		Entry l = p.left;
		p.left = l.right;
		if (l.right != null) l.right.parent = p;
			l.parent = p.parent;
		if (p.parent == null)
			root = l;
		else if (p.parent.right == p)
			p.parent.right = l;
		else p.parent.left = l;
			l.right = p;
		p.parent = l;
	}


	/** From CLR **/
	protected void fixAfterInsertion(Entry x)
	{
		x.color = RED;
		while (x != null && x != root && x.parent.color == RED)
		{
			if (parentOf(x) == leftOf(parentOf(parentOf(x))))
			{
	//System.out.print("  fix  ");
				Entry y = rightOf(parentOf(parentOf(x)));
				if (colorOf(y) == RED)
				{
	//System.out.print("  AfterInsertion  ");
				    setColor(parentOf(x), BLACK);
				    setColor(y, BLACK);
				    setColor(parentOf(parentOf(x)), RED);
				    x = parentOf(parentOf(x));
				}
				else
				{
	System.out.print("  BeforeInsertion  ");
					if (x == rightOf(parentOf(x))) 
					{
						x = parentOf(x);
						rotateLeft(x);
					}
					setColor(parentOf(x), BLACK);
					setColor(parentOf(parentOf(x)), RED);
					if (parentOf(parentOf(x)) != null) 
					rotateRight(parentOf(parentOf(x)));
				}
			}
			else
			{
	System.out.print("  unfix  ");
				Entry y = leftOf(parentOf(parentOf(x)));
				if (colorOf(y) == RED)
				{
	System.out.print("  AfterInsertion  ");
					setColor(parentOf(x), BLACK);
				 	setColor(y, BLACK);
					setColor(parentOf(parentOf(x)), RED);
					x = parentOf(parentOf(x));
				}
				else
				{
	System.out.print("  BeforeInsertion  ");
					if (x == leftOf(parentOf(x)))	
					{
						x = parentOf(x);
						rotateRight(x);
					}
					setColor(parentOf(x),  BLACK);
					setColor(parentOf(parentOf(x)), RED);
					if (parentOf(parentOf(x)) != null) 
						rotateLeft(parentOf(parentOf(x)));
				}
			}
		}
		root.color = BLACK;
	}


	protected void deleteEntry(Entry p)
	{
		decrementSize();
		if (p.left != null && p.right != null)
		{
			Entry s = successor(p);
			swapPosition(s, p);
		} 
		// Start fixup at replacement node, if it exists.
		Entry replacement = (p.left != null ? p.left : p.right);
		if (replacement != null)
		{
			// Link replacement to parent 
			replacement.parent = p.parent;
			if (p.parent == null)       
				root = replacement; 
			else if (p == p.parent.left)  
				p.parent.left  = replacement;
			else
				p.parent.right = replacement;
			// Null out links so they are OK to use by fixAfterDeletion.
			p.left = p.right = p.parent = null;      
			// Fix replacement
			if (p.color == BLACK) 
			fixAfterDeletion(replacement);
		} 
		else if (p.parent == null) // return if we are the only node.
	 		root = null;
		else //  No children. Use self as phantom replacement and unlink.
		{ 
			if (p.color == BLACK) 
			fixAfterDeletion(p);
			if (p.parent != null)
			{
				if (p == p.parent.left) 
				    p.parent.left = null;
				else if (p == p.parent.right) 
					p.parent.right = null;
				p.parent = null;
			}
		}
	}


    /** From CLR **/
	protected void fixAfterDeletion(Entry x)
	{
		while (x != root && colorOf(x) == BLACK)
		{
			if (x == leftOf(parentOf(x)))
			{
				Entry sib = rightOf(parentOf(x));
				if (colorOf(sib) == RED)
				{
					setColor(sib, BLACK);
					setColor(parentOf(x), RED);
					rotateLeft(parentOf(x));
					sib = rightOf(parentOf(x));
				}
				if (colorOf(leftOf(sib))  == BLACK && colorOf(rightOf(sib)) == BLACK)
				{
					setColor(sib,  RED);
					x = parentOf(x);
				} 
				else
				{
					if (colorOf(rightOf(sib)) == BLACK)
					{
						setColor(leftOf(sib), BLACK);
						setColor(sib, RED);
						rotateRight(sib);
						sib = rightOf(parentOf(x));
					}
					setColor(sib, colorOf(parentOf(x)));
					setColor(parentOf(x), BLACK);
					setColor(rightOf(sib), BLACK);
					rotateLeft(parentOf(x));
					x = root;
				}
			}
			else // symmetric
			{ 
				Entry sib = leftOf(parentOf(x));
				if (colorOf(sib) == RED)
				{
					setColor(sib, BLACK);
					setColor(parentOf(x), RED);
					rotateRight(parentOf(x));				
					sib = leftOf(parentOf(x));
				}
				if (colorOf(rightOf(sib)) == BLACK && colorOf(leftOf(sib)) == BLACK) 
				{
					setColor(sib,  RED);
					x = parentOf(x);
				} 
				else
				{
					if (colorOf(leftOf(sib)) == BLACK)
					{
						setColor(rightOf(sib), BLACK);
						setColor(sib, RED);
						rotateLeft(sib);
						sib = leftOf(parentOf(x));
					}
					setColor(sib, colorOf(parentOf(x)));
					setColor(parentOf(x), BLACK);
					setColor(leftOf(sib), BLACK);
					rotateRight(parentOf(x));
					x = root;
				}
			}
		}
		setColor(x, BLACK); 
	}


	protected void swapPosition(Entry x, Entry y)	
	{
		Entry px = x.parent, lx = x.left, rx = x.right;
		Entry py = y.parent, ly = y.left, ry = y.right;
		boolean xWasLeftChild = px != null && x == px.left;
		boolean yWasLeftChild = py != null && y == py.left;
		if (x == py)
		{  
			x.parent = y;
			if (yWasLeftChild)	{y.left = x; y.right = rx;} 
			else				{y.right = x; y.left = lx;}
		} 
		else
		{
			x.parent = py; 
	    		if (py != null)
			{
				if (yWasLeftChild)	py.left = x;
				else				py.right = x;
			}
			y.left = lx;  y.right = rx;
		}
		if (y == px)
		{
			y.parent = x;
			if (xWasLeftChild)	{x.left = y; x.right = ry;} 
			else				{x.right = y; x.left = ly;}
		}
		else
		{
			y.parent = px; 
			if (px != null)
			{
				if (xWasLeftChild)		px.left = y;
				else					px.right = y;
			}
			x.left = ly;   x.right = ry;  
		}
		if (x.left != null)	x.left.parent = x;
		if (x.right != null)	x.right.parent = x;
		if (y.left != null)	y.left.parent = y;
		if (y.right != null)	y.right.parent = y;
		boolean c = x.color;
		x.color = y.color; y.color = c;
		if (root == x)		root = y;
		else if (root == y)	root = x;
	}














    private static final long serialVersionUID = 919286545866124006L;
    private void writeObject(java.io.ObjectOutputStream s)
        throws java.io.IOException {
	// Write out the Comparator and any hidden stuff
	s.defaultWriteObject();

	// Write out size (number of Mappings)
	s.writeInt(size);

        // Write out keys and values (alternating)
	for (java.util.Iterator i = entrySet().iterator(); i.hasNext(); ) {
            Entry e = (Entry)i.next();
            s.writeObject(e.key);
            s.writeObject(e.value);
	}
    }
    private void readObject(final java.io.ObjectInputStream s)
        throws java.io.IOException, ClassNotFoundException {
	// Read in the Comparator and any hidden stuff
	s.defaultReadObject();

        // Read in size
        int size = s.readInt();

        buildFromSorted(size, null, s, null);
    }
    void readTreeSet(int size, java.io.ObjectInputStream s, Object defaultVal)
        throws java.io.IOException, ClassNotFoundException {
        buildFromSorted(size, null, s, defaultVal);
    }
    void addAllForTreeSet(SortedSet set, Object defaultVal) {
      try {
          buildFromSorted(set.size(), set.iterator(), null, defaultVal);
      } catch (java.io.IOException cannotHappen) {
      } catch (ClassNotFoundException cannotHappen) {
      }
    }


    /**
     * Linear time tree building algorithm from sorted data.  Can accept keys
     * and/or values from iterator or stream. This leads to too many
     * parameters, but seems better than alternatives.  The four formats
     * that this method accepts are:
     *
     *	  1) An iterator of Map.Entries.  (it != null, defaultVal == null).
     *    2) An iterator of keys.         (it != null, defaultVal != null).
     *	  3) A stream of alternating serialized keys and values.
     *					  (it == null, defaultVal == null).
     *	  4) A stream of serialized keys. (it == null, defaultVal != null).
     *
     * It is assumed that the comparator of the RobTreeMap is already set prior
     * to calling this method.
     *
     * @param size the number of keys (or key-value pairs) to be read from
     *	      the iterator or stream.
     * @param it If non-null, new entries are created from entries
     *        or keys read from this iterator.
     * @param it If non-null, new entries are created from keys and
     *	      possibly values read from this stream in serialized form.
     *        Exactly one of it and str should be non-null.
     * @param defaultVal if non-null, this default value is used for
     *        each value in the map.  If null, each value is read from
     *        iterator or stream, as described above.
     * @throws IOException propagated from stream reads. This cannot
     *         occur if str is null.
     * @throws ClassNotFoundException propagated from readObject. 
     *         This cannot occur if str is null.
     */
	private void buildFromSorted(int size, java.util.Iterator it, java.io.ObjectInputStream str, Object defaultVal) throws  java.io.IOException, ClassNotFoundException 
	{
		this.size = size;
		root = buildFromSorted(0, 0, size-1, computeRedLevel(size), it, str, defaultVal);
	}

    /**
     * Recursive "helper method" that does the real work of the
     * of the previous method.  Identically named parameters have
     * identical definitions.  Additional parameters are documented below.
     * It is assumed that the comparator and size fields of the RobTreeMap are
     * already set prior to calling this method.  (It ignores both fields.)
     *
     * @param level the current level of tree. Initial call should be 0.
     * @param lo the first element index of this subtree. Initial should be 0.
     * @param hi the last element index of this subtree.  Initial should be
     *	      size-1.
     * @param redLevel the level at which nodes should be red. 
     *        Must be equal to computeRedLevel for tree of this size.
     */
	private static Entry buildFromSorted(int level, int lo, int hi, int redLevel, java.util.Iterator it, java.io.ObjectInputStream str, Object defaultVal) throws  java.io.IOException, ClassNotFoundException
	{
        /*
         * Strategy: The root is the middlemost element. To get to it, we
         * have to first recursively construct the entire left subtree,
         * so as to grab all of its elements. We can then proceed with right
         * subtree. 
         *
         * The lo and hi arguments are the minimum and maximum
         * indices to pull out of the iterator or stream for current subtree.
         * They are not actually indexed, we just proceed sequentially,
         * ensuring that items are extracted in corresponding order.
         */
		if (hi < lo) return null;
		int mid = (lo + hi) / 2;        
		Entry left  = null;
		if (lo < mid) 
	            left = buildFromSorted(level+1, lo, mid - 1, redLevel, it, str, defaultVal);        
        // extract key and/or value from iterator or stream
		Object key;
	        Object value;
      	  if (it != null) { // use iterator
			if (defaultVal==null) {
				Map.Entry entry = (Map.Entry) it.next();
				key = entry.getKey();
                value = entry.getValue();
            } else {
                key = it.next();
                value = defaultVal;
            }
        } else { // use stream
            key = str.readObject();
            value = (defaultVal != null ? defaultVal : str.readObject());
        }

        Entry middle =  new Entry(key, value, null);
        
        // color nodes in non-full bottommost level red
        if (level == redLevel)
            middle.color = RED;
        
        if (left != null) { 
            middle.left = left; 
            left.parent = middle; 
        }
        
        if (mid < hi) {
            Entry right = buildFromSorted(level+1, mid+1, hi, redLevel,
                                          it, str, defaultVal);
            middle.right = right;
            right.parent = middle;
        }
        
        return middle;
    }

    /**
     * Find the level down to which to assign all nodes BLACK.  This is the
     * last `full' level of the complete binary tree produced by
     * buildTree. The remaining nodes are colored RED. (This makes a `nice'
     * set of color assignments wrt future insertions.) This level number is
     * computed by finding the number of splits needed to reach the zeroeth
     * node.  (The answer is ~lg(N), but in any case must be computed by same
     * quick O(lg(N)) loop.)
     */
    private static int computeRedLevel(int sz) {
        int level = 0;
        for (int m = sz - 1; m >= 0; m = m / 2 - 1) 
            level++;
        return level;
    }
}
